from abc import ABCMeta, abstractmethod
from typing import Generic, Iterable, Type, TypeVar

from typing_extensions import Self

_DataType = TypeVar('_DataType')


class AbstractDataModel(Generic[_DataType], metaclass=ABCMeta):

    @classmethod
    @abstractmethod
    def serialize(cls, data: _DataType) -> Self:
        ...

    @abstractmethod
    def deserialize(self) -> _DataType:
        ...

    @classmethod
    @abstractmethod
    def data_type(cls) -> Type[_DataType]:
        ...


def get_serializer(serializers: Iterable[Type[AbstractDataModel]], data_type: Type[_DataType]) -> Type[AbstractDataModel[_DataType]]:
    for serializer in serializers:
        if issubclass(data_type, serializer.data_type()):
            return serializer

    raise ValueError(f"No serializer found for {data_type}")



import os
import shutil
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import TypeVar

from typing_extensions import Self

_T = TypeVar('_T', bound='Serializable')


class Serializable(metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def load(cls, path: Path) -> Self:
        pass

    @abstractmethod
    def save(self, path: Path, *, rewrite: bool = False) -> None:
        pass

    @staticmethod
    def _create_empty_dir(directory: Path, *, clear: bool = False) -> None:
        if directory.exists():
            if not directory.is_dir():
                raise ValueError(f'Provided path {directory} is not a directory!')
            if clear:
                shutil.rmtree(directory)

        os.makedirs(directory, exist_ok=True)



from dataclasses import dataclass

from tp_interfaces.domain.abstract import AbstractNERCBasedType, NERCRegexp


@dataclass(frozen=True)
class NERCBasedType(AbstractNERCBasedType):
    _regexp: tuple[NERCRegexp, ...] = tuple()
    _black_regexp: tuple[NERCRegexp, ...] = tuple()
    _pretrained_nerc_models: tuple[str, ...] = tuple()
    _dictionary: tuple[str, ...] = tuple()
    _black_list: tuple[str, ...] = tuple()

    def __post_init__(self):
        if self._regexp:
            object.__setattr__(self, '_regexp', tuple(map(self._convert_to_regexp, self._regexp)))
        if self._black_regexp:
            object.__setattr__(self, '_black_regexp', tuple(map(self._convert_to_regexp, self._black_regexp)))

        for attr in ['_regexp', '_black_regexp', '_pretrained_nerc_models', '_dictionary', '_black_list']:
            if isinstance(getattr(self, attr), list):
                object.__setattr__(self, attr, tuple(getattr(self, attr)))

    @property
    async def regexp(self) -> tuple[NERCRegexp, ...]:
        return self._regexp

    @property
    async def black_regexp(self) -> tuple[NERCRegexp, ...]:
        return self._black_regexp

    @property
    async def pretrained_nerc_models(self) -> tuple[str, ...]:
        return self._pretrained_nerc_models

    @property
    async def dictionary(self) -> tuple[str, ...]:
        return self._dictionary

    @property
    async def black_list(self) -> tuple[str, ...]:
        return self._black_list

    @staticmethod
    def _convert_to_regexp(obj) -> NERCRegexp:
        if isinstance(obj, NERCRegexp):
            return obj
        if isinstance(obj, dict):
            return NERCRegexp(**obj)
        raise ValueError



from typing import AsyncContextManager, Protocol

from typing_extensions import Self


class AsyncModel(AsyncContextManager[Self], Protocol):
    async def __aenter__(self) -> Self:
        return self



from typing import Tuple


def get_correct_indices(indices: str, max_count: int = 0) -> Tuple[int, ...]:
    """
    Convert arbitrary table or column indices from a string (e.g., "1, 3, 5-8, 10-end") to
    a tuple of correct integer indexes (e.g., (1, 3, 5, 6, 7, 8, 10, ...)).
    """
    correct_indices = []
    for item in indices.replace(" ", "").split(","):
        index_range = item.split("-")
        count = len(index_range)
        if count == 2:
            start, end = map(str, index_range)
            if start.isnumeric() and end.isnumeric():
                for index in range(int(start) - 1, int(end)):
                    correct_indices.append(index)
            if start.isnumeric() and end == "end":
                for index in range(int(start) - 1, max_count):
                    correct_indices.append(index)
        if count == 1 and index_range[0].isnumeric():
            correct_indices.append(int(index_range[0]) - 1)
    return tuple(correct_indices)



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Optional

from tdm.abstract.datamodel.domain import AbstractDomainType


@dataclass(frozen=True)
class NERCRegexp(object):
    regexp: str
    context_regexp: Optional[str] = None
    auto_create: bool = False


class AbstractNERCBasedType(AbstractDomainType, metaclass=ABCMeta):

    @property
    @abstractmethod
    async def regexp(self) -> tuple[NERCRegexp, ...]:
        pass

    @property
    @abstractmethod
    async def black_regexp(self) -> tuple[NERCRegexp, ...]:
        pass

    @property
    @abstractmethod
    async def pretrained_nerc_models(self) -> tuple[str, ...]:
        pass

    @property
    @abstractmethod
    async def dictionary(self) -> tuple[str, ...]:
        pass

    @property
    @abstractmethod
    async def black_list(self) -> tuple[str, ...]:
        pass



from tdm.abstract.json_schema import generate_model
from tdm.datamodel.domain import ConceptType
from tdm.datamodel.domain.types import DocumentType

generate_model(label='concept')(ConceptType)
generate_model(label='document')(DocumentType)



import asyncio


async def async_task_executor(tasks) -> tuple:
    if not tasks:
        return tuple()
    _, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_EXCEPTION)
    for task in pending:
        task.cancel()
    res = []
    for task in tasks:
        if task.exception():
            raise task.exception()
        else:
            res.append(task.result())
    return tuple(res)



from abc import ABCMeta, abstractmethod
from typing import AsyncContextManager, Iterable

from tdm.abstract.datamodel import AbstractDomain
from typing_extensions import Self


class AbstractDomainChangeHook(object):
    @abstractmethod
    def __call__(self, domain: AbstractDomain) -> None:
        pass

    @classmethod
    @abstractmethod
    def from_config(cls, config: dict) -> Self:
        pass


class DomainProducer(AsyncContextManager, metaclass=ABCMeta):
    def __init__(self, hooks: Iterable[AbstractDomainChangeHook] = tuple()):
        self._hooks: list[AbstractDomainChangeHook] = list(hooks)

    @abstractmethod
    async def has_changed(self) -> bool:
        pass

    async def get_domain(self) -> AbstractDomain:
        domain = await self._get_domain()
        for hook in self._hooks:
            hook(domain)
        return domain

    @abstractmethod
    async def _get_domain(self) -> AbstractDomain:
        pass

    def register_hook(self, hook: AbstractDomainChangeHook) -> None:
        self._hooks.append(hook)

    @classmethod
    @abstractmethod
    def from_config(cls, config: dict) -> Self:
        pass



__all__ = [
    'Serializable', 'DictWrapper', 'DirectorySerializableMixin', 'load_object', 'PicklableMixin'
]

from .abstract import Serializable
from .dict_wrapper import DictWrapper
from .directory_mixin import DirectorySerializableMixin
from .manipulations import load_object
from .pickle_mixin import PicklableMixin



from operator import attrgetter
from typing import Generic, Iterable, Optional, Tuple, Type, TypeVar

from pydantic import create_model

from tp_interfaces.abstract import AbstractDocumentProcessor, AbstractUpdatableModel, AbstractUpdate, UpdateMode
from tp_interfaces.processors.composite import SequentialDocumentProcessor

_Update = TypeVar('_Update', bound=AbstractUpdate)


class SequentialUpdate(AbstractUpdate):
    mode: Optional[UpdateMode] = None


class UpdatableSequentialDocumentProcessor(AbstractUpdatableModel[_Update], SequentialDocumentProcessor, Generic[_Update]):

    def __init__(self, processors: Iterable[Tuple[str, AbstractDocumentProcessor]]):
        super().__init__(processors)

        update_types = {}
        update_config_extractors = []

        for model_name, model in zip(self._model_names, self._models):
            if not isinstance(model, AbstractUpdatableModel):
                continue
            if model_name in update_types:
                raise ValueError(f"duplicate model name {model_name}")
            update_types[model_name] = (model.update_type, None)
            update_config_extractors.append(attrgetter(model_name))

        if len(update_types) == 1:
            self._update_type = update_types.popitem()[1][0]
            self._update_config_extractors = (lambda config: config,)
        elif any(not name for name in update_types):
            raise ValueError(f"empty model name couldn't be processed")
        else:
            self._update_type = create_model('RuntimeSequentialUpdate', **update_types, __base__=SequentialUpdate)
            self._update_config_extractors = tuple(update_config_extractors)

    @property
    def update_type(self) -> Type[_Update]:
        return self._update_type

    async def update(self, update: _Update) -> None:
        filtered = filter(lambda x: isinstance(x, AbstractUpdatableModel), self._models)
        for processor, update_config_extractor in zip(filtered, self._update_config_extractors):
            sub_update = update_config_extractor(update)

            if sub_update:
                await processor.update(sub_update)

    async def _add(self, update: _Update) -> None:
        raise NotImplementedError

    async def _remove(self, update: _Update) -> None:
        raise NotImplementedError

    @classmethod
    def build(cls, processors: Iterable[Tuple[str, AbstractDocumentProcessor]], *,
              merge: bool = True) -> AbstractDocumentProcessor:
        processors = tuple(cls._merge_models(processors)) if merge else tuple(processors)
        if len(processors) == 1:
            return processors[0][1]

        if any(isinstance(processor[1], AbstractUpdatableModel) for processor in processors):
            return UpdatableSequentialDocumentProcessor(processors)

        return SequentialDocumentProcessor(processors)



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Generic, Optional, TypeVar

from .processor import AbstractDocumentProcessor

DEFAULT_CACHE_DIR = Path('trainer_cache')


@dataclass
class TrainingResults:
    model: AbstractDocumentProcessor
    metrics: Optional[Dict[str, float]]


_Processor = TypeVar('_Processor', bound=AbstractDocumentProcessor)
_TrainingResults = TypeVar('_TrainingResults', bound=TrainingResults)


class AbstractTrainer(Generic[_Processor, _TrainingResults], metaclass=ABCMeta):

    @classmethod
    @abstractmethod
    def from_config(cls, config: dict):
        pass

    @abstractmethod
    def train(self, cache_dir: Optional[Path] = DEFAULT_CACHE_DIR, pretrained_model: Optional[_Processor] = None) -> _TrainingResults:
        pass



from abc import ABCMeta, abstractmethod
from collections import defaultdict
from typing import Any, Callable, Coroutine, Dict, Iterable, List, Optional, Sequence, Tuple

from tp_interfaces.abstract import ImmutableBaseModel

ConfigurableMessageProcessor = Callable[[Sequence[Any], ImmutableBaseModel], Tuple[Any, ...]]
AsyncConfigurableMessageProcessor = Callable[[Sequence[Any], ImmutableBaseModel], Coroutine]


class AbstractModelInput(ImmutableBaseModel, metaclass=ABCMeta):
    @abstractmethod
    def get_message(self) -> Any:
        pass

    @abstractmethod
    def get_config(self) -> Optional[ImmutableBaseModel]:
        pass


def batch_process_inputs(inputs: Sequence[AbstractModelInput], processor: ConfigurableMessageProcessor) \
        -> Tuple[Any, ...]:
    config2messages: Dict[ImmutableBaseModel, List[Tuple[int, Any]]] = defaultdict(list)
    for i, inp in enumerate(inputs):  # group messages by configs
        config2messages[inp.get_config()].append((i, inp.get_message()))

    outputs: List[Any] = [None] * len(inputs)  # prepare output to preserve messages order
    for config, messages_with_indicies in config2messages.items():  # batch process strings with response order guaranteed
        indices, messages = zip(*messages_with_indicies)
        indices: Iterable[int]
        for idx, output in zip(indices, processor(messages, config)):
            outputs[idx] = output

    return tuple(outputs)


async def async_batch_process_inputs(inputs: Sequence[AbstractModelInput], processor: AsyncConfigurableMessageProcessor) \
        -> Tuple[Any, ...]:
    config2messages: Dict[ImmutableBaseModel, List[Tuple[int, Any]]] = defaultdict(list)
    for i, inp in enumerate(inputs):  # group messages by configs
        config2messages[inp.get_config()].append((i, inp.get_message()))

    outputs: List[Any] = [None] * len(inputs)  # prepare output to preserve messages order
    for config, messages_with_indicies in config2messages.items():  # batch process strings with response order guaranteed
        indices, messages = zip(*messages_with_indicies)
        indices: Iterable[int]

        for idx, output in zip(indices, await processor(messages, config)):
            outputs[idx] = output

    return tuple(outputs)



__all__ = [
    'ComponentValueType', 'ConceptType', 'DocumentType', 'NERCBasedType', 'IdentifyingPropertyType', 'PropertyType', 'RelationPropertyType',
    'RelationType', 'AtomValueType', 'CompositeValueType'
]

from ._component import ComponentValueType
from ._concept import ConceptType, DocumentType
from ._nerc import NERCBasedType
from ._property import IdentifyingPropertyType, PropertyType, RelationPropertyType
from ._relation import RelationType
from ._value import AtomValueType, CompositeValueType



import time
import uuid
from dataclasses import dataclass
from typing import Protocol, Type, TypeVar, runtime_checkable

from typing_extensions import Self

from tp_interfaces.abstract.processor import AbstractProcessor
from tp_interfaces.abstract.schema import ImmutableBaseModel

_Config = TypeVar('_Config', bound=ImmutableBaseModel)


@dataclass(frozen=True, init=False)
class File:
    path: str
    filename: str
    checksum: str

    def __init__(self, path: str, filename: str, checksum: str, **kwargs):
        object.__setattr__(self, "path", path)
        object.__setattr__(self, "filename", filename)
        object.__setattr__(self, "checksum", checksum)
        for key, value in kwargs.items():
            object.__setattr__(self, key, value)


@dataclass(frozen=True, init=False)
class Message:
    id: str
    timestamp: int
    file: File | None
    parent_uuid: str | None

    def __init__(
            self,
            id_: str | None = None,
            file: File | None = None,
            parent_uuid: str | None = None,
            timestamp: int | None = None,
            **kwargs
    ):
        object.__setattr__(self, "id", id_ if id_ else str(uuid.uuid4()))
        object.__setattr__(self, "timestamp", timestamp if timestamp else int(time.time()))
        object.__setattr__(self, "file", file)
        object.__setattr__(self, "parent_uuid", parent_uuid)
        for key, value in kwargs.items():
            object.__setattr__(self, key, value)

    @classmethod
    def create(
            cls,
            parent_uuid: str | None = None,
            s3_path: str | None = None,
            file_hash: str | None = None,
            file_name: str | None = None,
            file_fields: dict | None = None,
            fields: dict | None = None
    ) -> Self:
        file_fields = {} if file_fields is None else file_fields
        fields = {} if fields is None else fields
        file = File(path=s3_path, checksum=file_hash, filename=file_name, **file_fields) if s3_path else None
        message = Message(parent_uuid=parent_uuid, file=file, **fields)
        return message


@runtime_checkable
class AbstractMessageProcessor(AbstractProcessor[_Config, Message, Message], Protocol[_Config]):

    @property
    def config_type(self) -> Type[_Config]:
        raise NotImplementedError

    @property
    def input_type(self) -> Type[Message]:
        return Message

    @property
    def output_type(self) -> Type[Message]:
        return Message



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Optional

from tdm.abstract.datamodel.domain import AbstractDomainType


@dataclass(frozen=True)
class RelExtModel(object):
    relation_type: str
    invert_direction: bool = False
    source_annotation: Optional[str] = None
    target_annotation: Optional[str] = None


class AbstractRelExtBasedType(AbstractDomainType, metaclass=ABCMeta):
    @property
    @abstractmethod
    async def pretrained_relext_models(self) -> tuple[RelExtModel, ...]:
        pass



from abc import ABCMeta
from typing import Generic, TypeVar

from tp_interfaces.abstract.model import AbstractModelWrapper
from .model import AbstractUpdatableModel, UpdatableModelMixin
from .update import AbstractUpdate

_Update = TypeVar('_Update', bound=AbstractUpdate)


class UpdatableModelWrapper(
    AbstractModelWrapper[AbstractUpdatableModel[_Update]],
    UpdatableModelMixin[_Update], Generic[_Update],
    metaclass=ABCMeta
):
    """
    Utility class that creates either updatable extractor wrapper (if wrapped model is updatable) or general extractor wrapper.
    Note: this class subclasses should not override `from_config` method (AbstractModelWrapper `from_config` should be used)
    """

    def __init__(self, model: AbstractUpdatableModel[_Update], *args, **kwargs):
        AbstractModelWrapper[AbstractUpdatableModel[_Update]].__init__(self, model)
        if not isinstance(model, AbstractUpdatableModel):
            raise ValueError
        UpdatableModelMixin.__init__(self, model)

    @classmethod
    def from_config(cls, model: AbstractUpdatableModel, config: dict):
        if isinstance(model, AbstractUpdatableModel):
            kwargs = cls.parse_config(config)
            return cls(model, **kwargs)
        return cls.__base__.from_config(model, config)



from abc import ABCMeta, abstractmethod
from typing import Generic, TypeVar

from typing_extensions import Self

from tp_interfaces.abstract.model import AsyncModel

_Model = TypeVar('_Model', bound=AsyncModel)


class AbstractModelWrapper(AsyncModel, Generic[_Model], metaclass=ABCMeta):
    def __init__(self, model: _Model, **kwargs):
        self._model = model

    async def __aenter__(self) -> Self:
        await self._model.__aenter__()
        return self

    async def __aexit__(self, *exc):
        await self._model.__aexit__(*exc)

    @classmethod
    def from_config(cls, model: _Model, config: dict) -> Self:
        kwargs = cls.parse_config(config)
        return cls(model, **kwargs)

    @staticmethod
    @abstractmethod
    def parse_config(config: dict) -> dict:
        pass



import inspect
from contextvars import ContextVar
from functools import wraps
from typing import Callable, ContextManager, Optional

_LOG_EXTRAS = ContextVar[dict]('_LOG_EXTRAS')


def get_log_extras() -> dict:
    try:
        return _LOG_EXTRAS.get()
    except LookupError:
        return {}


class LogExtrasManager(ContextManager):
    __slots__ = ('_extras', '_token')

    def __init__(self, extras: dict):
        self._extras = extras
        self._token = None

    def __enter__(self) -> None:
        self._token = _LOG_EXTRAS.set(self._extras)

    def __exit__(self, *exc) -> Optional[bool]:
        _LOG_EXTRAS.reset(self._token)
        return super().__exit__(*exc)


def update_log_extras(*, __to_str__: bool = True, **kwargs) -> LogExtrasManager:
    extras = dict(get_log_extras())
    extras.update({k: str(v) if __to_str__ else v for k, v in kwargs.items()})
    return LogExtrasManager(extras)


def with_log_extras(**params):
    """
    params is a mapping from extras key to constant value or to callable (called for function params)
    """

    constant_params = {}
    callable_params = {}

    for key, value in params.items():
        (callable_params if isinstance(value, Callable) else constant_params)[key] = value

    def decorator(f):
        @wraps(f)
        def wrap(*args, **kwargs):
            signature = inspect.signature(f)
            binding = signature.bind(*args, **kwargs)
            binding.apply_defaults()
            extras = {k: v(binding.arguments) for k, v in callable_params.items()}
            with update_log_extras(**constant_params, **extras):
                return f(*args, **kwargs)

        return wrap

    return decorator



from dataclasses import dataclass

from tdm.abstract.json_schema import generate_model

from tp_interfaces.domain.abstract import AbstractComponentValueType
from ._relext import RelExtBasedType
from ._value import AtomValueType

__deps = (AtomValueType,)


@generate_model(label='component')
@dataclass(frozen=True)
class ComponentValueType(RelExtBasedType, AbstractComponentValueType):
    pass



from typing import AsyncContextManager, Generic, Iterable, TypeVar

from typing_extensions import Self

from .model import AsyncModel

_AsyncModel = TypeVar('_AsyncModel', bound=AsyncContextManager)


class AbstractAsyncCompositeModel(AsyncModel, Generic[_AsyncModel]):
    def __init__(self, models: Iterable[_AsyncModel]):
        self._models = tuple(models)

    async def __aenter__(self) -> Self:
        for m in self._models:
            await m.__aenter__()
        return self

    async def __aexit__(self, __exc_type, __exc_value, __traceback):
        for m in self._models[::-1]:
            await m.__aexit__(__exc_type, __exc_value, __traceback)



from pathlib import Path
from typing import Type, TypeVar, Union

from tp_interfaces.serializable.abstract import Serializable
from tp_interfaces.serializable.directory_mixin import DirectorySerializableMixin
from tp_interfaces.serializable.pickle_mixin import PicklableMixin

_Serializable = TypeVar('_Serializable', bound=Union[Serializable, PicklableMixin])


def load_object(path: Path, *, expected_class: Type[_Serializable] = Serializable) -> _Serializable:
    if path.is_dir():
        model = DirectorySerializableMixin.load(path)
    else:
        model = PicklableMixin.load(path)
    if not isinstance(model, expected_class):
        raise TypeError(f'Expected {expected_class.__name__} at {path} but loaded {model.__class__.__name__}!')
    return model



import inspect
import json
import time
from abc import ABCMeta
from importlib import import_module
from pathlib import Path
from typing import Any, Callable, Optional

from typing_extensions import Self

from tp_interfaces.serializable.abstract import Serializable

INFO_FILE = 'info.json'
CONFIG_FILE = 'config.json'


def load_info_file(directory: Path, *, default: Optional[dict] = None) -> dict:
    info_file = directory / INFO_FILE

    if info_file.exists():
        with info_file.open('r', encoding='utf-8') as f:
            return json.load(f)

    if default is None:
        raise ValueError(f'File {INFO_FILE} is missing from {directory} and no default value is provided for info!')

    return default


def _get_constructor_param_values(target_type: type):
    """
    get target type constructor parameter values

    @param target_type: type of the created object
    @return: mapping from parameter name to its value (except self)
    """
    frame = inspect.currentframe().f_back  # get caller frame
    caller_class = None
    names = tuple(inspect.signature(target_type.__init__).parameters)
    # unlovely replacement for inspect.f_code.co_qualname below:
    while caller_class is not target_type:
        frame = frame.f_back
        current_local_types = {type(v) for v in frame.f_locals.values()}
        if target_type in current_local_types and frame.f_code.co_names[1] == "__init__" \
                and set(names).issubset(set(frame.f_locals.keys())):
            caller_class = type(frame.f_locals['self'])

    return {
        name: frame.f_locals[name] for name in names[1:]  # skip self
    }


class DirectorySerializableMixin(Serializable, metaclass=ABCMeta):
    """
    Mixin class that implements basic directory serialization and deserialization.
    This mixin should be initialized in serializable object constructor to collect actual parameters.

    This mixin stores initial constructor parameter values and serializes them.
    Object will be created with the same parameters while deserialization.

    Make sure that superclass constructors do not mutate their arguments and that there is no name collision
    in the variables of child and parent classes which descend from this mixin. An illustration of these failure modes
    can be seen in tp_interfaces_tests/serializable/serializable.py
    """

    CREATED_INFO_FIELD = 'created'
    CLASS_INFO_FIELD = 'class'
    EXTERNAL_PARAMETERS = 'external'

    def __init__(self):
        params = _get_constructor_param_values(type(self))
        self._inline_config: dict[str, Any] = {}
        self._serializable: dict[str, tuple[Any, Callable, Callable]] = {}
        for name, value in params.items():
            try:
                # check if class specifies special method for current value serialization/deserialization
                serializer, deserializer = self._serializer(value)
                self._serializable[name] = value, serializer, deserializer
            except ValueError:
                # if not, assume value is json-serializable
                self._inline_config[name] = value

    @classmethod
    def load(cls, path: Path) -> Self:
        if not path.exists():
            raise ValueError(f'Path {path} does not exist!')

        if not path.is_dir():
            raise ValueError(f'Provided path {path} is not a directory!')

        info = load_info_file(path)
        class_name: str = info.get(cls.CLASS_INFO_FIELD)
        if class_name is None:
            raise ValueError(f'Broken info file: there is no {cls.CLASS_INFO_FIELD} field value')
        try:
            module, name = class_name.rsplit('.', 1)
            module = import_module(module)
            cls_ = getattr(module, name)
        except Exception as e:
            raise ImportError(f"Couldn't import {class_name} class", e)
        if not isinstance(cls_, type):
            raise TypeError(f"{class_name} is not a class")
        if not issubclass(cls_, cls):
            raise TypeError(f"Expected class {cls.__name__} at {path} but found {cls_.__name__}!")

        return cls_._load_from_directory(path, info=info)

    @classmethod
    def _load_from_directory(cls, directory_path: Path, info: dict) -> Self:
        with (directory_path / CONFIG_FILE).open('r', encoding='utf-8') as f:
            params = json.load(f)
        if not isinstance(params, dict):
            raise ValueError
        params: dict
        for param, deserializer in info.get(cls.EXTERNAL_PARAMETERS, {}).items():
            module = import_module(deserializer['module'])
            if deserializer.get('class') is not None:
                module = getattr(module, deserializer['class'])
            func = getattr(module, deserializer['name'])
            params[param] = func(directory_path / param)
        return cls(**params)

    def _get_info(self) -> dict:
        def build_config(deserializer: dict | Callable) -> dict:
            if isinstance(deserializer, dict):
                return deserializer
            qualname = deserializer.__qualname__
            if '.' in qualname:
                class_name, name = qualname.rsplit('.', 1)
            else:
                class_name, name = None, qualname
            return {
                "module": inspect.getmodule(deserializer).__name__,
                "class": class_name,
                "name": name
            }

        return {
            self.CREATED_INFO_FIELD: time.strftime("%Y-%m-%d %H:%M"),
            self.CLASS_INFO_FIELD: f'{self.__class__.__module__}.{self.__class__.__name__}',
            self.EXTERNAL_PARAMETERS: {
                name: build_config(deserializer) for name, (_, _, deserializer) in self._serializable.items()
            },
            **self._get_extra_info()
        }

    def _get_extra_info(self) -> dict:
        return {}

    def save(self, path: Path, *, rewrite: bool = False) -> None:
        self._create_empty_dir(path, clear=rewrite)

        info_file = path / INFO_FILE
        info = self._get_info()
        with info_file.open('w', encoding='utf-8') as f:
            json.dump(info, f, ensure_ascii=False, indent=2)
        return self._save_to_directory(path)

    def _save_to_directory(self, directory_path: Path) -> None:
        with (directory_path / CONFIG_FILE).open('w', encoding='utf-8') as f:
            json.dump(self._inline_config, f, ensure_ascii=False, indent=2)
        for param, (value, serializer, _) in self._serializable.items():
            serializer(value, directory_path / param)

    @classmethod
    def _serializer(cls, obj) -> tuple[Callable[[Any, Path], None], Callable[[Path], Any] | dict]:
        """
        get special serialization/deserialization methods for object

        @param obj: object to be serialized
        @return: serialization function and deserialization function (or its specification: module, class and method name)
        """
        if isinstance(obj, Serializable):
            return lambda o, p: o.save(p), obj.load
        from tp_interfaces.serializable import DictWrapper
        if isinstance(obj, DictWrapper):
            return DictWrapper.save, DictWrapper.load
        raise ValueError



from dataclasses import dataclass

from tdm.abstract.json_schema import generate_model

from tp_interfaces.domain.abstract import AbstractRelationType
from ._concept import ConceptType
from ._relext import RelExtBasedType

__deps = (ConceptType,)


@generate_model(label='relation')
@dataclass(frozen=True)
class RelationType(RelExtBasedType, AbstractRelationType):

    def __post_init__(self):
        RelExtBasedType.__post_init__(self)
        AbstractRelationType.__post_init__(self)



import logging
import operator
from operator import attrgetter
from typing import Iterable, Sequence, Type, TypeVar

from pydantic import BaseModel, ValidationError, create_model
from tdm import TalismanDocument

from tp_interfaces.abstract import AbstractDocumentProcessor, ImmutableBaseModel
from tp_interfaces.abstract.model.composite import AbstractAsyncCompositeModel
from tp_interfaces.abstract.model.merge import MergeModel
from tp_interfaces.logging.time import AsyncTimeMeasurer

logger = logging.getLogger(__name__)

_Config = TypeVar('_Config', bound=BaseModel)


class SequentialConfig(ImmutableBaseModel):
    pass


class SequentialDocumentProcessor(
    AbstractAsyncCompositeModel[AbstractDocumentProcessor],
    AbstractDocumentProcessor[SequentialConfig]
):

    def __init__(self, processors: Iterable[tuple[str, AbstractDocumentProcessor]]):
        processors = tuple(processors)
        AbstractAsyncCompositeModel[AbstractDocumentProcessor].__init__(self, map(operator.itemgetter(1), processors))
        AbstractDocumentProcessor.__init__(self)

        self._config_type = None
        self._config_extractors = ()
        self._model_names = tuple(map(operator.itemgetter(0), processors))

    async def __aenter__(self):
        await AbstractAsyncCompositeModel.__aenter__(self)

        config_extractors = []
        config_types = {}
        nonempty_idx = None
        for idx, (model_name, model) in enumerate(zip(self._model_names, self._models)):
            if not model.config_type.model_fields:
                config = model.config_type()
                config_extractors.append(lambda _, config=config: config)  # use hack (config=config) to fix Late Binding Closures
                continue
            if model_name in config_types:
                raise ValueError(f"duplicate model name {model_name}")
            try:
                default_config = model.config_type()
            except ValidationError:
                default_config = ...
            config_types[model_name] = (model.config_type, default_config)
            config_extractors.append(attrgetter(model_name))
            nonempty_idx = idx

        if len(config_types) == 1:
            config_extractors[nonempty_idx] = lambda config: config
            self._config_type = next(iter(config_types.values()))[0]
        else:
            if any(not name for name in config_types):
                raise ValueError(f"empty model name couldn't be processed")
            self._config_type = create_model('RuntimeSequentialConfig', **config_types, __base__=SequentialConfig)

        self._config_extractors = tuple(config_extractors)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await AbstractAsyncCompositeModel.__aexit__(self, exc_type, exc_val, exc_tb)
        self._config_type = None
        self._config_extractors = ()

    async def process_doc(self, document: TalismanDocument, config: SequentialConfig) -> TalismanDocument:
        return (await self.process_docs([document], config))[0]

    async def process_docs(
            self,
            documents: Sequence[TalismanDocument],
            config: SequentialConfig
    ) -> tuple[TalismanDocument, ...]:
        for processor, config_extractor in zip(self._models, self._config_extractors):
            async with AsyncTimeMeasurer(f"processing {len(documents)} documents with {processor}", logger=logger):
                documents = await processor.process_docs(documents, config_extractor(config))
        return documents

    @property
    def config_type(self) -> Type[SequentialConfig]:
        if self._config_type is None:
            raise RuntimeError('First call the __aenter__ method!')
        return self._config_type

    @classmethod
    def build(cls, processors: Iterable[tuple[str, AbstractDocumentProcessor]], *, merge: bool = True) -> AbstractDocumentProcessor:
        processors = tuple(cls._merge_models(processors)) if merge else tuple(processors)
        if len(processors) == 1:
            return processors[0][1]
        return SequentialDocumentProcessor(processors)

    @staticmethod
    def _merge_models(models: Iterable[tuple[str, AbstractDocumentProcessor]]) -> Iterable[tuple[str, AbstractDocumentProcessor]]:
        current_merger: MergeModel | None = None
        merger_name: str | None = None
        for name, model in models:
            if current_merger is not None:
                if current_merger.can_be_merged(model):
                    current_merger = current_merger.merge(model)
                else:
                    yield merger_name, current_merger
                    merger_name, current_merger = None, None
            if current_merger is None:
                if isinstance(model, MergeModel):
                    merger_name, current_merger = name, model
                else:
                    yield name, model
        if current_merger is not None:
            yield merger_name, current_merger



from tdm.abstract.datamodel import AbstractDomain
from tdm.datamodel.domain import set_default_domain
from typing_extensions import Self

from tp_interfaces.domain.interfaces import AbstractDomainChangeHook


class SetDefaultDomainHook(AbstractDomainChangeHook):
    def __call__(self, domain: AbstractDomain) -> None:
        set_default_domain(domain)

    @classmethod
    def from_config(cls, config: dict) -> Self:
        return cls()



from tdm.abstract.datamodel import AbstractDomain
from tdm.datamodel.domain import Domain
from typing_extensions import Self

from tp_interfaces.domain.interfaces import DomainProducer
from tp_interfaces.domain.model import DomainModel

EMPTY_DOMAIN = Domain(())


class StubDomainProducer(DomainProducer):
    def __init__(self, domain: AbstractDomain = EMPTY_DOMAIN):
        super().__init__()
        self._domain = domain

    async def __aenter__(self):
        return self

    async def __aexit__(self, *exc):
        pass

    async def has_changed(self) -> bool:
        return False

    async def _get_domain(self) -> AbstractDomain:
        return self._domain

    @classmethod
    def from_config(cls, config: dict) -> Self:
        """Config example:
        {
            "domain": {
                ...`KBSchema` config...
            }
        }
        """
        return cls(domain=DomainModel.model_validate(config['domain']).deserialize())



from abc import ABCMeta
from dataclasses import dataclass
from typing import Optional

from tdm.datamodel.domain import AtomValueType

from ._nerc import AbstractNERCBasedType


@dataclass(frozen=True)
class AbstractLiteralValueType(AbstractNERCBasedType, AtomValueType, metaclass=ABCMeta):
    value_restriction: Optional[tuple[str, ...]] = None

    def __post_init__(self):
        AtomValueType.__post_init__(self)
        if isinstance(self.value_restriction, list):
            object.__setattr__(self, 'value_restriction', tuple(self.value_restriction))



from abc import ABCMeta, abstractmethod
from typing import Protocol, TypeVar, runtime_checkable

from typing_extensions import Self

from .model import AsyncModel

_Model = TypeVar('_Model', bound=AsyncModel, covariant=True)


@runtime_checkable
class MergeModel(Protocol[_Model]):

    @abstractmethod
    def can_be_merged(self, model: _Model) -> bool:
        ...

    @abstractmethod
    def merge(self, model: _Model) -> Self:
        ...


class IDBasedMergeModel(MergeModel, metaclass=ABCMeta):

    def can_be_merged(self, model: AsyncModel) -> bool:
        if isinstance(model, IDBasedMergeModel):
            return self.id == model.id
        return False

    @property
    @abstractmethod
    def id(self) -> str:
        pass



from argparse import ArgumentTypeError


# From https://stackoverflow.com/questions/15008758/parsing-boolean-values-with-argparse
def string_to_bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise ArgumentTypeError(f"Truthy value expected: got {v} but expected one of yes/no, true/false, t/f, y/n, 1/0 (case insensitive).")



__all__ = [
    'AbstractProcessor',
    'AbstractCategoryAwareNodeProcessor',
    'AbstractMessageProcessor', 'Message',
    'AbstractNodeProcessor',
    'AbstractDocumentProcessor',
    'AbstractTrainer'
]

from .abstract import AbstractProcessor
from .category import AbstractCategoryAwareNodeProcessor
from .message import AbstractMessageProcessor, Message
from .node import AbstractNodeProcessor
from .processor import AbstractDocumentProcessor
from .trainer import AbstractTrainer



from dataclasses import dataclass

from tdm.abstract.json_schema import generate_model

from tp_interfaces.domain.abstract import AbstractIdentifyingPropertyType, AbstractPropertyType, AbstractRelationPropertyType
from ._concept import ConceptType
from ._nerc import NERCBasedType
from ._relation import RelationType
from ._relext import RelExtBasedType
from ._value import AtomValueType

__deps = (AtomValueType, ConceptType, RelationType)


@generate_model(label='property')
@dataclass(frozen=True)
class PropertyType(RelExtBasedType, AbstractPropertyType):

    def __post_init__(self):
        RelExtBasedType.__post_init__(self)
        AbstractPropertyType.__post_init__(self)


@generate_model(label='id_property')
@dataclass(frozen=True)
class IdentifyingPropertyType(NERCBasedType, AbstractIdentifyingPropertyType):

    def __post_init__(self):
        NERCBasedType.__post_init__(self)
        AbstractIdentifyingPropertyType.__post_init__(self)


@generate_model(label='r_property')
@dataclass(frozen=True)
class RelationPropertyType(RelExtBasedType, AbstractRelationPropertyType):
    def __post_init__(self):
        RelExtBasedType.__post_init__(self)
        AbstractRelationPropertyType.__post_init__(self)



class WrongRequest(Exception):  # noqa N818
    """
    Created for cases when error caused not by error in service code, but by error in request
    Currently usage the only usage is config validation for document processors
    """


class ProcessorConfigError(WrongRequest):
    pass



from abc import ABCMeta
from dataclasses import dataclass, field

from tdm.abstract.datamodel import FactStatus
from tdm.datamodel.domain import PropertyType, RelationPropertyType
from tdm.datamodel.facts import ConceptFact, PropertyFact, ValueFact

from ._nerc import AbstractNERCBasedType
from ._relext import AbstractRelExtBasedType


@dataclass(frozen=True)
class AbstractPropertyType(AbstractRelExtBasedType, PropertyType, metaclass=ABCMeta):
    isIdentifying: bool = field(init=False, default=False)  # noqa N815

    def build_fact(self, status: FactStatus, source: ConceptFact, target: ValueFact) -> PropertyFact:
        return PropertyFact(status, self, source, target)


@dataclass(frozen=True)
class AbstractIdentifyingPropertyType(AbstractNERCBasedType, PropertyType, metaclass=ABCMeta):
    isIdentifying: bool = field(init=False, default=True)  # noqa N815


class AbstractRelationPropertyType(AbstractRelExtBasedType, RelationPropertyType, metaclass=ABCMeta):
    pass



import asyncio
from abc import ABC, ABCMeta, abstractmethod
from pathlib import Path
from typing import Any, AsyncIterator, Dict, Iterator, Type, TypeVar

from tdm import TalismanDocument

from tp_interfaces.helpers.io import check_path_existence

_AbstractConfigurableReader = TypeVar('_AbstractConfigurableReader', bound='AbstractConfigurableReader')


class AbstractPathConstructor(metaclass=ABCMeta):
    @abstractmethod
    def get_data_paths(self, filepath: Path) -> Iterator[Any]:
        pass


class OneFilePathConstructor(AbstractPathConstructor):
    def get_data_paths(self, filepath: Path) -> Iterator[Dict[str, Path]]:
        yield {'filepath': Path(filepath)}


class MultiFilePathConstructor(AbstractPathConstructor):
    def __init__(self, pattern: str = "*"):
        self._pattern = pattern

    def get_data_paths(self, filepath: Path) -> Iterator[Dict[str, Path]]:
        check_path_existence(filepath)
        for path_doc in filepath.glob(self._pattern):
            yield {'filepath': Path(path_doc)}


class _FakeAsyncIterator(AsyncIterator[TalismanDocument]):

    def __init__(self, sync_iterator: Iterator[TalismanDocument]):
        self._iterator = sync_iterator

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            await asyncio.sleep(0)  # switch coroutine
            return next(self._iterator)
        except StopIteration:
            raise StopAsyncIteration


class AbstractReader(metaclass=ABCMeta):
    def __init__(self, filepath: Path):
        self._filepath = filepath

    @abstractmethod
    def read_doc(self, *args, **kwargs) -> Iterator[TalismanDocument]:
        pass

    async def aread_doc(self, *args, **kwargs) -> AsyncIterator[TalismanDocument]:
        return _FakeAsyncIterator(self.read_doc(*args, **kwargs))  # TODO: implement async read methods

    def read(self) -> Iterator[TalismanDocument]:
        for path_config in self.path_constructor.get_data_paths(self._filepath):
            for doc in self.read_doc(**path_config):
                yield doc

    async def aread(self) -> AsyncIterator[TalismanDocument]:
        for path_config in self.path_constructor.get_data_paths(self._filepath):
            async for doc in await self.aread_doc(**path_config):
                yield doc

    @property
    def path_constructor(self) -> AbstractPathConstructor:
        return OneFilePathConstructor()


class AbstractConfigurableReader(AbstractReader, metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def from_config(cls: _AbstractConfigurableReader, config) -> _AbstractConfigurableReader:
        pass


def replace_path_constructor(reader_type: Type[AbstractReader], path_constructor) -> Type[AbstractReader]:
    class WrappedReaderClass(reader_type, ABC):
        @property
        def path_constructor(self) -> AbstractPathConstructor:
            return path_constructor

    return WrappedReaderClass



from abc import ABCMeta

from tdm.abstract.datamodel import FactStatus
from tdm.datamodel.domain import RelationType
from tdm.datamodel.facts import ConceptFact, RelationFact

from ._relext import AbstractRelExtBasedType


class AbstractRelationType(AbstractRelExtBasedType, RelationType, metaclass=ABCMeta):
    def build_fact(self, status: FactStatus, source: ConceptFact, target: ConceptFact) -> RelationFact:
        return RelationFact(status, self, source, target)



__all__ = [
    'AbstractComponentValueType',
    'AbstractNERCBasedType', 'NERCRegexp',
    'AbstractIdentifyingPropertyType', 'AbstractPropertyType', 'AbstractRelationPropertyType',
    'AbstractRelationType',
    'AbstractRelExtBasedType', 'RelExtModel',
    'AbstractLiteralValueType'
]

from ._component import AbstractComponentValueType
from ._nerc import AbstractNERCBasedType, NERCRegexp
from ._property import AbstractIdentifyingPropertyType, AbstractPropertyType, AbstractRelationPropertyType
from ._relation import AbstractRelationType
from ._relext import AbstractRelExtBasedType, RelExtModel
from ._value import AbstractLiteralValueType



from abc import ABCMeta
from typing import Protocol, Type, TypeVar, runtime_checkable

from tdm import TalismanDocument

from tp_interfaces.abstract.processor import AbstractProcessor
from tp_interfaces.abstract.schema import ImmutableBaseModel
from tp_interfaces.serializable import Serializable

_Config = TypeVar('_Config', bound=ImmutableBaseModel)


@runtime_checkable
class AbstractDocumentProcessor(AbstractProcessor[_Config, TalismanDocument, TalismanDocument], Protocol[_Config]):

    @property
    def config_type(self) -> Type[_Config]:
        raise NotImplementedError

    @property
    def input_type(self) -> Type[TalismanDocument]:
        return TalismanDocument

    @property
    def output_type(self) -> Type[TalismanDocument]:
        return TalismanDocument


class AbstractSerializableDocumentProcessor(AbstractDocumentProcessor, Serializable, metaclass=ABCMeta):
    pass



import os
import tarfile
import tempfile
from io import BytesIO
from pathlib import Path
from typing import Callable, Dict, IO, TypeVar, Union


def create_tar_wrapper(names2paths: Dict[str, Union[Path, str]]) -> bytes:
    """
    Function tars single file with arcname "model" and return bytes buffer with it
    """
    if not names2paths:
        raise ValueError("No directories given for tar!")

    bytes_buffer = BytesIO()
    with tarfile.open(mode="w:", fileobj=bytes_buffer) as tar:
        for name, path in names2paths.items():
            if isinstance(path, str):
                path = Path(path)
            if not os.path.isdir(path):
                raise NotADirectoryError(f'Path \'{path}\' not a directory')

            tar.add(path, arcname=name)
    return bytes_buffer.getvalue()


_T = TypeVar('_T')


def extract_tar(tared_model: IO[bytes], extractor: Callable[[Path], _T]) -> _T:
    """
    Extracts all tarred files into temporary directory and applies extractor
    """
    with tarfile.open("r:", fileobj=tared_model) as tar:
        with tempfile.TemporaryDirectory() as dir_name:
            tar.extractall(dir_name)
            return extractor(Path(dir_name))



import asyncio
from functools import partial
from typing import AsyncIterator, Protocol, Sequence, Tuple, Type, TypeVar, runtime_checkable

from tp_interfaces.abstract.model import AsyncModel
from tp_interfaces.abstract.schema import ImmutableBaseModel
from tp_interfaces.helpers.task_execution import async_task_executor
from tp_interfaces.logging.context import with_log_extras

_Config = TypeVar('_Config', bound=ImmutableBaseModel)
_Input = TypeVar('_Input')
_Output = TypeVar('_Output')


@runtime_checkable
class AbstractProcessor(AsyncModel, Protocol[_Config, _Input, _Output]):
    async def process_doc(self, document: _Input, config: _Config) -> _Output:
        pass

    async def process_docs(self, documents: Sequence[_Input], config: _Config) -> Tuple[_Output, ...]:
        log_extras = with_log_extras(doc_id=lambda kwargs: kwargs['document'].id)
        wrapped_processor = log_extras(self.process_doc)
        configured_processor = partial(wrapped_processor, config=config)
        coroutines = map(configured_processor, documents)
        tasks = [asyncio.create_task(coro) for coro in coroutines]
        return await async_task_executor(tasks)

    async def process_stream(
            self,
            documents: AsyncIterator[_Input],
            config: _Config,
            batch_size: int,
            concurrency: int = 1,
            results_queue_max_size: int = 0
    ) -> AsyncIterator[_Output]:
        """
        Asynchronously processes a stream of documents in batches and yields results as they become available.

        This method consumes an asynchronous stream of `_Input` objects, groups them into batches of the specified size,
        and processes the batches concurrently using `process_docs` method. Results are streamed back as they are processed.

        Notes:
        - Output order is **not guaranteed** to match input order due to concurrent batch processing.
        - Exceptions during processing are skipped silently.
        - Processing begins immediately and continues as input documents are streamed in.

        :param documents: asynchronous stream of input documents
        :param config: configuration object for processing (same for all document batches)
        :param batch_size: number of documents per batch
        :param concurrency: number of concurrent workers. Defaults to 1
        :param results_queue_max_size: maximum size of the results queue. Defaults to 0 (unbounded)
        :return: asynchronous iterator over processed documents
        """

        batches_queue: asyncio.Queue[Sequence[_Input]] = asyncio.Queue(maxsize=concurrency)
        results_queue: asyncio.Queue[_Output | Exception | None] = asyncio.Queue(maxsize=results_queue_max_size)

        async def batcher():
            batch = []
            async for doc in documents:
                batch.append(doc)
                if len(batch) >= batch_size:
                    await batches_queue.put(tuple(batch))
                    batch = []
            if batch:
                await batches_queue.put(tuple(batch))

        async def worker():
            while True:
                batch = await batches_queue.get()
                try:
                    result = await self.process_docs(batch, config)
                    for doc in result:
                        await results_queue.put(doc)
                except Exception as e:
                    await results_queue.put(e)
                finally:
                    batches_queue.task_done()  # mark batch processed after adding results in queue

        batcher_task = asyncio.create_task(batcher())
        worker_tasks = [asyncio.create_task(worker()) for _ in range(concurrency)]

        async def finalizer():
            await batcher_task  # no more batches will be added to batches queue
            await batches_queue.join()  # all added tasks are processed
            for worker_task in worker_tasks:
                worker_task.cancel()  # interrupt workers (they are blocked with input_queue.get method)
            await asyncio.gather(*worker_tasks, return_exceptions=True)  # all workers now awaited
            await results_queue.put(None)  # end of results signal

        asyncio.create_task(finalizer())

        while True:
            result = await results_queue.get()
            if result is None:
                break
            if isinstance(result, Exception):
                continue  # TODO: add logging
            yield result

    @property
    def config_type(self) -> Type[_Config]:
        raise NotImplementedError

    @property
    def input_type(self) -> Type[_Input]:
        raise NotImplementedError

    @property
    def output_type(self) -> Type[_Output]:
        raise NotImplementedError



__all__ = [
    'ModelTypeFactory',
    'AbstractAsyncCompositeModel', 'AbstractConfigConstructableModel', 'AbstractModelWrapper',
    'AbstractProcessor', 'AbstractCategoryAwareNodeProcessor', 'AbstractDocumentProcessor',
    'AbstractMessageProcessor', 'Message',
    'AbstractNodeProcessor', 'AbstractTrainer',
    'ImmutableBaseModel',
    'AbstractUpdatableModel', 'AbstractUpdate', 'UpdatableModelWrapper', 'UpdateMode'
]

from .configuration import ModelTypeFactory
from .model import AbstractAsyncCompositeModel, AbstractConfigConstructableModel, AbstractModelWrapper
from .processor import AbstractCategoryAwareNodeProcessor, AbstractDocumentProcessor, AbstractMessageProcessor, AbstractNodeProcessor, \
    AbstractProcessor, AbstractTrainer, Message
from .schema import ImmutableBaseModel
from .updatable import AbstractUpdatableModel, AbstractUpdate, UpdatableModelWrapper, UpdateMode



from abc import ABCMeta
from dataclasses import dataclass
from functools import singledispatch

from tdm.abstract.datamodel import AbstractValue
from tdm.datamodel.values import Coordinates, Date, DateTimeValue, DoubleValue, GeoPointValue, IntValue, StringValue, Time, TimestampValue


@dataclass(frozen=True)
class AbstractFilter(metaclass=ABCMeta):  # noqa B024: this is actually abstract class
    def __post_init__(self):
        if type(self) is AbstractFilter:
            raise TypeError("Can't instantiate abstract class")


@singledispatch
def get_filters(value: AbstractValue, **kwargs) -> AbstractFilter:
    raise NotImplementedError


@dataclass(frozen=True)
class _StringFilter:
    str: str
    exact: bool = False

    def __post_init__(self):
        object.__setattr__(self, 'str', self.str.replace('"', '?'))  # see TKL-987


@dataclass(frozen=True)
class StringFilter(AbstractFilter):
    stringFilter: _StringFilter  # noqa: N815


@get_filters.register
def _string_filter(value: StringValue, exact: bool = False) -> StringFilter:
    return StringFilter(_StringFilter(value.value, exact))


@dataclass(frozen=True)
class _IntFilter:
    start: int
    end: int


@dataclass(frozen=True)
class IntFilter(AbstractFilter):
    intFilter: _IntFilter  # noqa N815


def _int_timestamp_filter(value: IntValue | TimestampValue, **kwargs) -> IntFilter:
    return IntFilter(_IntFilter(value.value, value.value))


@get_filters.register
def _int_filter(value: IntValue, **kwargs) -> IntFilter:
    return _int_timestamp_filter(value)


@get_filters.register
def _timestamp_filter(value: TimestampValue, **kwargs) -> IntFilter:
    return _int_timestamp_filter(value)


@dataclass(frozen=True)
class _DoubleFilter:
    start: float
    end: float


@dataclass(frozen=True)
class DoubleFilter(AbstractFilter):
    doubleFilter: _DoubleFilter  # noqa N815


@get_filters.register
def _double_filter(value: DoubleValue, **kwargs) -> DoubleFilter:
    return DoubleFilter(_DoubleFilter(value.value, value.value))


@dataclass(frozen=True)
class _DateTime:
    date: Date
    time: Time | None = None


@dataclass(frozen=True)
class _DateTimeFilter:
    start: _DateTime
    end: _DateTime


@dataclass(frozen=True)
class DateTimeFilter(AbstractFilter):
    dateTimeFilter: _DateTimeFilter  # noqa N815


@get_filters.register
def _date_time_filter(value: DateTimeValue, **kwargs) -> DateTimeFilter:
    date_time = _DateTime(value.date, value.time)
    return DateTimeFilter(_DateTimeFilter(date_time, date_time))


@dataclass(frozen=True)
class _GeoFilter:
    point: Coordinates | None
    name: str | None
    radius: float


@dataclass(frozen=True)
class GeoFilter(AbstractFilter):
    geoFilter: _GeoFilter  # noqa N815


@get_filters.register
def _geo_filter(value: GeoPointValue, **kwargs) -> GeoFilter:
    return GeoFilter(_GeoFilter(value.point, value.name, 0.0001))



import inspect
from functools import wraps
from pathlib import Path

from typing_extensions import Self

from tp_interfaces.serializable.abstract import Serializable
from tp_interfaces.serializable.manipulations import load_object


class DictWrapper(dict[str, Serializable]):
    def __init__(self, content: dict[str, Serializable]):
        dict.__init__(self, content)

    @classmethod
    def load(cls, path: Path) -> Self:
        content = {}
        for file in path.glob("*"):
            content[file.name] = load_object(file)
        return cls(content)

    def save(self, path: Path, *, rewrite: bool = False) -> None:
        for name, item in self.items():
            item.save(path / name)

    @classmethod
    def convert_dict(cls, param: str):
        def wrapper(f):
            signature = inspect.signature(f)

            @wraps(f)
            def wrap(*args, **kwargs):
                bound_args = signature.bind(*args, **kwargs)
                bound_args.apply_defaults()
                named_args = bound_args.arguments
                named_args[param] = cls(named_args[param])
                return f(**named_args)

            return wrap

        return wrapper



import logging
import os
from multiprocessing.pool import ThreadPool
from typing import Iterable

import requests

from talisman_tools.helper.serializers import MessageModel
from tp_interfaces.abstract import Message

logger = logging.getLogger(__name__)


class TControllerUploader:

    def __init__(self) -> None:
        self.tcontroller_url = os.environ.get("TCONTROLLER_URL")
        self.num_processes = int(os.getenv("NUM_PROCESSES_UPLOAD", 8))

        logger.warning(f"Use TControllerUploader (url: {self.tcontroller_url})")

    def upload(self, messages: Iterable[Message], topic: str) -> None:
        worker = lambda message: self.upload_message(message, topic)
        with ThreadPool(processes=self.num_processes) as pool:
            pool.map(worker, messages)

    def upload_message(self, message: Message, topic: str) -> None:
        message = MessageModel.serialize(message)
        data = {
            "id": message.id,
            "topic": topic,
            "message": message.model_dump(by_alias=True, exclude_none=True)
        }

        status_code = 500
        while status_code != 200:
            r = requests.post(f"{self.tcontroller_url}/messages/add", json=data)
            status_code = r.status_code
            logger.warning(f"upload message to tcontroller with uuid {data['id']}, code: {status_code}, content: {r.content.decode()}")



from tdm.abstract.datamodel import AbstractDomainType


def get_id(type_id: str | AbstractDomainType) -> str:
    """
    Utility function that extracts str type_id.
    """
    if isinstance(type_id, str):
        return type_id
    return type_id.id  # noqa: AbstractDomainType is EnsureIdentifiable



from typing import Protocol, TypeVar

_Model = TypeVar('_Model', covariant=True)


class AbstractConfigConstructableModel(Protocol[_Model]):
    @classmethod
    def from_config(cls, config: dict) -> _Model:
        ...



from asyncio import Lock
from typing import Optional

from tdm.abstract.datamodel import AbstractDomain

from tp_interfaces.domain.interfaces import DomainProducer


class _Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


lock = Lock()


class DomainManager(metaclass=_Singleton):
    def __init__(self):
        self._factory: Optional[DomainProducer] = None
        self._domain = None

        self._entered = 0

    def __bool__(self):
        return self._factory is not None

    def set_producer(self, domain_factory: DomainProducer) -> None:
        if self._entered:
            raise AttributeError("Domain producer is already in use")
        self._factory = domain_factory

    def del_producer(self):
        if self._entered:
            raise AttributeError("Domain producer is already in use")
        self._factory = None

    async def __aenter__(self):
        if self._factory is None:
            raise ValueError("No domain producer is set")
        async with lock:
            if not self._entered:
                await self._factory.__aenter__()
                self._domain = await self._factory.get_domain()
            self._entered += 1
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        async with lock:
            self._entered -= 1
            if not self._entered:
                await self._factory.__aexit__(exc_type, exc_val, exc_tb)
                self._domain = None

    @property
    async def has_changed(self) -> bool:
        async with lock:
            return await self._has_changed()

    @property
    async def domain(self) -> AbstractDomain:
        async with lock:
            has_changed = await self._has_changed()
            while has_changed:
                self._domain = await self._factory.get_domain()
                has_changed = await self._has_changed()
            return self._domain

    async def _has_changed(self) -> bool:
        if not self._entered:
            raise AttributeError
        return await self._factory.has_changed()



from abc import ABCMeta
from collections import defaultdict
from typing import Callable, Generic, Iterable, Iterator, Optional, Type, TypeVar

from tdm.abstract.datamodel import AbstractNode

from tp_interfaces.abstract.schema import ImmutableBaseModel
from .node import AbstractNodeProcessor
from .processor import AbstractDocumentProcessor
from .. import AbstractAsyncCompositeModel

_Config = TypeVar('_Config', bound=ImmutableBaseModel)
_Category = TypeVar('_Category')

_Node = TypeVar('_Node', bound=AbstractNode)


class AbstractCategoryAwareNodeProcessor(
    AbstractAsyncCompositeModel[AbstractNodeProcessor[_Config, _Node]],
    AbstractNodeProcessor[_Config, _Node],
    Generic[_Config, _Node, _Category],
    metaclass=ABCMeta
):
    def __init__(
            self,
            category2processor: dict[_Category, AbstractNodeProcessor[_Config, _Node]],
            categorizer: Callable[[_Node], _Category],
            default_processor: Optional[AbstractDocumentProcessor[_Config]] = None
    ):
        self._category2processor = dict(category2processor)
        super().__init__(self._category2processor.values())
        self._models: tuple[AbstractDocumentProcessor[_Config], ...]
        self._categorizer = categorizer
        self._default = default_processor

    def _process_node(self, node: _Node, config: _Config) -> _Node:
        return next(self._process_nodes([node], config))

    def _process_nodes(self, nodes: Iterable[_Node], config: _Config) -> Iterator[_Node]:
        category2node = defaultdict(list)

        for node in nodes:
            category = self._categorizer(node)
            category2node[category].append(node)

        for category, nodes in category2node.items():
            processor = self._category2processor.get(category, self._default)
            if processor is None:
                raise ValueError(f"No processor registered for {category} (available categories: {list(self._category2processor)})")
            yield from processor._process_nodes(nodes, config)

    @property
    def config_type(self) -> Type[_Config]:
        return self._models[0].config_type



from typing import Type

from tdm.abstract.datamodel import AbstractDomainType
from tdm.abstract.json_schema import AbstractLabeledModel, ModelsGenerator, get_model_generator
from tdm.abstract.json_schema.decorator import _MODELS_GENERATORS


def register_domain_types_models() -> Type[AbstractLabeledModel]:
    _MODELS_GENERATORS[AbstractDomainType] = ModelsGenerator(AbstractDomainType)
    import tp_interfaces.domain.model.types as types
    types

    # TODO: here plugin for extra domain types

    return get_model_generator(AbstractDomainType).generate_labeled_model('DomainTypesModel')


DomainTypesModel = register_domain_types_models()



import hashlib


def calculate_md5sum(file_path: str) -> str:
    file_hash = hashlib.md5()
    with open(file_path, "rb") as file:
        chunk = file.read(8192)
        while chunk:
            file_hash.update(chunk)
            chunk = file.read(8192)
    return str(file_hash.hexdigest())



from abc import ABCMeta

from tdm.datamodel.domain import ComponentValueType

from ._relext import AbstractRelExtBasedType


class AbstractComponentValueType(AbstractRelExtBasedType, ComponentValueType, metaclass=ABCMeta):
    pass



from typing import Type

from tp_interfaces.abstract import ModelTypeFactory
from tp_interfaces.domain.interfaces import AbstractDomainChangeHook


def _get_update_tdm_hook() -> Type[AbstractDomainChangeHook]:
    from .tdm_factory import UpdateTDMDefaultDomainHook
    return UpdateTDMDefaultDomainHook


def _get_set_default_domain_hook() -> Type[AbstractDomainChangeHook]:
    from .default_domain import SetDefaultDomainHook
    return SetDefaultDomainHook


DOMAIN_CHANGE_HOOKS = ModelTypeFactory({
    'tdm': _get_update_tdm_hook,
    'default': _get_set_default_domain_hook
})



__all__ = [
    'AbstractAsyncCompositeModel',
    'AbstractConfigConstructableModel',
    'AsyncModel',
    'AbstractModelWrapper',
]

from .composite import AbstractAsyncCompositeModel
from .constructable import AbstractConfigConstructableModel
from .model import AsyncModel
from .wrapper import AbstractModelWrapper



from functools import singledispatch
from typing import Iterator

from tdm import TalismanDocument
from tdm.abstract.datamodel import AbstractFact
from tdm.datamodel.facts import AtomValueFact, ComponentFact, CompositeValueFact, ConceptFact, MentionFact, PropertyFact, RelationFact


def get_document_facts(doc: TalismanDocument) -> Iterator[AbstractFact]:
    """Yield all information about document fact(with id = doc.id)"""
    try:
        doc_fact = doc.get_fact(doc.id)
    except KeyError:
        return

    yield doc_fact
    for prop in doc.get_facts(PropertyFact, filter_=PropertyFact.source_filter(ConceptFact.id_filter(doc.id))):
        yield prop
        yield from get_value(prop.target, doc)

    for relation in doc.get_facts(RelationFact, filter_=RelationFact.source_filter(ConceptFact.id_filter(doc.id))):
        yield relation
        yield relation.target


@singledispatch
def get_value(value: AbstractFact, doc: TalismanDocument) -> Iterator[AbstractFact]:
    """Recursively retrieves all dependencies for a value fact (mentions, components)."""
    raise NotImplementedError


@get_value.register
def _(value: AtomValueFact, doc: TalismanDocument) -> Iterator[AbstractFact]:
    yield value
    yield from doc.related_facts(value, MentionFact)


@get_value.register
def _(value: CompositeValueFact, doc: TalismanDocument) -> Iterator[AbstractFact]:
    yield value
    for component in doc.related_facts(value, ComponentFact):
        yield component
        yield from get_value(component.target, doc)



import pickle
from pathlib import Path
from typing import TypeVar

from typing_extensions import Self

from tp_interfaces.serializable.abstract import Serializable

_T = TypeVar('_T', bound='PicklableMixin')


def read_dvc_bytes(path: Path) -> bytes:
    import dvc.api
    # problem with dvc.api.open: https://github.com/iterative/dvc/issues/4667
    return dvc.api.read(str(path.with_suffix('')), mode='rb')


class PicklableMixin(Serializable):

    @classmethod
    def load(cls, path: Path) -> Self:
        if path.suffix == '.dvc':
            model = pickle.loads(read_dvc_bytes(path))
        else:
            with path.open('rb') as f:
                model = pickle.load(f)
        if not isinstance(model, cls):
            raise pickle.PickleError(f"Pickled object at {path} is not an instance of {cls.__name__} class.")
        return model

    def save(self, path: Path, *, rewrite: bool = False) -> None:
        if path.exists() and not rewrite:
            raise Exception(f"Saving path exists: {path}")
        self._create_empty_dir(path.parent)

        with path.open('wb') as f:
            pickle.dump(self, f, protocol=5)  # fixed protocol version to avoid issues with serialization on Python 3.10+ versions



from dataclasses import dataclass

from tdm.abstract.json_schema import generate_model
from tdm.datamodel.domain import CompositeValueType

from tp_interfaces.domain.abstract import AbstractLiteralValueType
from ._nerc import NERCBasedType


@generate_model(label='atom')
@dataclass(frozen=True)
class AtomValueType(NERCBasedType, AbstractLiteralValueType):

    def __post_init__(self):
        NERCBasedType.__post_init__(self)
        AbstractLiteralValueType.__post_init__(self)


generate_model(label='composite')(CompositeValueType)



from typing import Type

from tp_interfaces.domain.interfaces import DomainProducer


def _get_stub() -> Type[DomainProducer]:
    from tp_interfaces.domain.stub import StubDomainProducer
    return StubDomainProducer


DOMAIN_FACTORY = _get_stub()



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("VERSION", "r", encoding="utf-8") as f:
    version = f.read()

setup(
    name='talisman-interfaces',
    version=version,
    description='Talisman Processor base interfaces',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='ISPRAS Talisman NLP team',
    author_email='modis@ispras.ru',
    maintainer='Vladimir Mayorov',
    maintainer_email='vmayorov@ispras.ru',
    packages=find_packages(include=['tp_interfaces', 'tp_interfaces.*']),
    install_requires=[
        'talisman-dm>=1.3.1,<2',
        'pydantic~=2.5', 'more-itertools>=8.12.0'
    ],
    extras_require={
        'dvc': ['dvc[s3]>=3.33.3']
    },
    data_files=[('', ['VERSION'])],
    python_requires='>=3.10',
    license='Apache Software License',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: Apache Software License'
    ]
)



from pydantic import BaseModel
from tdm.abstract.datamodel import AbstractDomain
from tdm.datamodel.domain import Domain

from ._types_registry import DomainTypesModel


class DomainModel(BaseModel):
    types: DomainTypesModel

    def deserialize(self) -> Domain:
        return Domain(self.types.deserialize({}))

    @classmethod
    def serialize(cls, domain: AbstractDomain) -> 'DomainModel':
        return cls(
            types=DomainTypesModel.serialize(domain.types)
        )



from abc import ABCMeta, abstractmethod
from typing import Generic, Type, TypeVar

from .update import AbstractUpdate, UpdateMode
from ..model import AsyncModel

_Update = TypeVar('_Update', bound=AbstractUpdate)


class AbstractUpdatableModel(AsyncModel, Generic[_Update]):

    async def update(self, update: _Update) -> None:
        if update.mode is UpdateMode.add:
            await self._add(update)
        else:
            await self._remove(update)

    @abstractmethod
    async def _add(self, update: _Update) -> None:
        pass

    @abstractmethod
    async def _remove(self, update: _Update) -> None:
        pass

    @property
    @abstractmethod
    def update_type(self) -> Type[_Update]:
        pass


class UpdatableModelMixin(AbstractUpdatableModel[_Update], Generic[_Update], metaclass=ABCMeta):
    def __init__(self, model: AbstractUpdatableModel):
        self._model = model

    async def update(self, update: _Update) -> None:
        await self._model.update(update)

    async def _add(self, update: _Update) -> None:
        await self._model._add(update)

    async def _remove(self, update: _Update) -> None:
        await self._model._remove(update)

    @property
    def update_type(self) -> Type[_Update]:
        return self._model.update_type



from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import AsyncIterator, Iterable, TextIO

from tdm import TalismanDocument
from typing_extensions import Protocol


class AsyncTextIO(Protocol):
    @property
    def closed(self) -> bool:
        ...

    async def writable(self) -> bool:
        ...

    async def write(self, b: str, /) -> int:
        ...


class AbstractSerializer(metaclass=ABCMeta):
    @abstractmethod
    def serialize(self, doc: TalismanDocument, stream: TextIO):
        pass

    @abstractmethod
    async def aserialize(self, doc: TalismanDocument, stream: AsyncTextIO):
        pass

    @staticmethod
    def _check_stream(stream: TextIO):
        if stream.closed or not stream.writable():
            raise Exception("stream  is closed or not writeable")

    @staticmethod
    async def _a_check_stream(stream: AsyncTextIO):
        if stream.closed or not await stream.writable():
            raise Exception("stream  is closed or not writeable")


class AbstractPathSerializer(metaclass=ABCMeta):
    @abstractmethod
    def serialize(self, docs: Iterable[TalismanDocument], path: Path):
        pass

    @abstractmethod
    async def aserialize(self, docs: AsyncIterator[TalismanDocument], path: Path):
        pass



from collections import defaultdict
from typing import Iterable

from tdm.abstract.datamodel import AbstractFact, AbstractNode, AbstractNodeMention, FactStatus
from tdm.datamodel.domain import AtomValueType, ConceptType, PropertyType, RelationType
from tdm.datamodel.facts import AtomValueFact, ConceptFact, MentionFact, PropertyFact, RelationFact
from tdm.datamodel.mentions import TextNodeMention
from tdm.datamodel.nodes import TableCellNode, TableCellNodeMetadata, TableNode, TableRowNode, TextNode


def generate_concept_facts(
        cpt_type: str | ConceptType,
        props: dict[str | PropertyType, tuple[str | AtomValueType, TextNode | AbstractNodeMention]],
        relations: dict[str | RelationType, Iterable[AbstractFact]] = None,
        status: FactStatus = FactStatus.NEW
) -> list[AbstractFact]:
    if relations is None:
        relations = {}
    result = []
    cpt = ConceptFact(status, cpt_type)
    result.append(cpt)
    for prop_type, (value_type, value_mention) in props.items():
        value, mention = generate_atom_value_facts(value_type, value_mention, status)
        result.extend([value, mention, PropertyFact(status, prop_type, cpt, value)])
    for rel_type, (target, *other) in relations.items():
        result.append(target)
        result.append(RelationFact(status, rel_type, cpt, target))
        result.extend(other)
    return result


def generate_atom_value_facts(
        value_type: str | AtomValueType, mention: TextNode | AbstractNodeMention, status: FactStatus = FactStatus.NEW
) -> tuple[AtomValueFact, MentionFact]:
    if isinstance(mention, TextNode):
        mention = TextNodeMention(mention, 0, len(mention.content))
    value = AtomValueFact(status, value_type)
    mention = MentionFact(status, mention, value)
    return value, mention


def generate_table_nodes(
        table: Iterable[Iterable[str]],
        table_id: str = 'table',
        vertical: bool = True,
        header_numbers: tuple[int, ...] = (0,)
) -> tuple[TableNode, dict[AbstractNode, Iterable[AbstractNode]], tuple[tuple[TextNode, ...], ...]]:
    root: TableNode = TableNode(id=table_id)
    structure: dict[AbstractNode, list[AbstractNode]] = defaultdict(list)
    cells: list[list[TextNode]] = []

    for i, row in enumerate(table):
        row_node = TableRowNode(id=f'{table_id}-{i}')
        structure[root].append(row_node)
        row_nodes = []
        for j, cell in enumerate(row):
            cell_node = TableCellNode(
                id=f'{table_id}-{i}-{j}',
                metadata=TableCellNodeMetadata(header=((i if vertical else j) in header_numbers))
            )
            structure[row_node].append(cell_node)
            text_node = TextNode(id=f'{table_id}-{i}-{j}-content', content=cell)
            structure[cell_node].append(text_node)
            if vertical or j != 0:
                row_nodes.append(text_node)
        if i != 0 or not vertical:
            cells.append(row_nodes)

    return root, structure, tuple(tuple(row) for row in cells)



from dataclasses import dataclass

from tp_interfaces.domain.abstract import AbstractRelExtBasedType, RelExtModel


@dataclass(frozen=True)
class RelExtBasedType(AbstractRelExtBasedType):
    _pretrained_relext_models: tuple[RelExtModel, ...] = tuple()

    def __post_init__(self):
        if self._pretrained_relext_models:
            object.__setattr__(self, '_pretrained_relext_models', tuple(map(self._convert_to_model, self._pretrained_relext_models)))
        elif isinstance(self._pretrained_relext_models, list):
            object.__setattr__(self, '_pretrained_relext_models', tuple(self._pretrained_relext_models))

    @property
    async def pretrained_relext_models(self) -> tuple[RelExtModel, ...]:
        return self._pretrained_relext_models

    @staticmethod
    def _convert_to_model(obj) -> RelExtModel:
        if isinstance(obj, RelExtModel):
            return obj
        if isinstance(obj, dict):
            return RelExtModel(**obj)
        raise ValueError



import enum

from tdm.datamodel.values import DateTimeValue, DoubleValue, GeoPointValue, IntValue, LinkValue, StringValue, TimestampValue


class ValueTypes(enum.Enum):
    Date = DateTimeValue
    Double = DoubleValue
    Geo = GeoPointValue
    Int = IntValue
    Timestamp = TimestampValue
    Link = LinkValue
    String = StringValue

    @classmethod
    def get(cls, key: str, default_value=None):
        if key in cls.__members__.keys():
            return cls[key].value
        return default_value



from abc import ABCMeta, abstractmethod
from functools import partial
from typing import Generic, Iterable, Iterator, Type, TypeVar

from tdm import TalismanDocument
from tdm.abstract.datamodel import AbstractNode

from tp_interfaces.abstract.schema import ImmutableBaseModel
from .processor import AbstractDocumentProcessor

_Config = TypeVar('_Config', bound=ImmutableBaseModel)
_Node = TypeVar('_Node', bound=AbstractNode)


class AbstractNodeProcessor(AbstractDocumentProcessor, Generic[_Config, _Node], metaclass=ABCMeta):
    """
    A base abstract processor for processing document nodes.
    """

    async def process_doc(self, document: TalismanDocument, config: _Config) -> TalismanDocument:
        filter_node = partial(self._filter_node, document=document)
        nodes = document.get_nodes(type_=self._node_type, filter_=filter_node)
        return document.with_nodes(self._process_nodes(nodes, config=config))

    def _filter_node(self, node: _Node, document: TalismanDocument) -> bool:
        return True

    @property
    def _node_type(self) -> Type[_Node]:
        return AbstractNode

    @abstractmethod
    def _process_node(self, node: _Node, config: _Config) -> _Node:
        pass

    def _process_nodes(self, nodes: Iterable[_Node], config: _Config) -> Iterator[_Node]:
        return map(partial(self._process_node, config=config), nodes)



import asyncio
import logging
import time
import types
from abc import ABCMeta
from functools import wraps
from logging import Logger
from typing import Any, AsyncContextManager, ContextManager

_LOGGER = logging.getLogger(__name__)


def is_async_callable(f: Any):
    if not callable(f):
        return False
    if isinstance(f, types.FunctionType):
        return asyncio.iscoroutinefunction(f)
    if hasattr(f, '__call__'):  # noqa: B004 I want to check __call__ attribute, not to test if f is callable
        return asyncio.iscoroutinefunction(f.__call__)


class AbstractTimeMeasurer(metaclass=ABCMeta):  # noqa B024: this is an abstract class
    def __init__(
            self,
            message: str, *,
            inline_time: bool = False,
            logger: Logger = _LOGGER,
            warning_threshold: float = None,
            extra: dict[str, Any] = None
    ):
        self._message = message
        self._inline_time = inline_time
        self._logger = logger
        self._warning_threshold = warning_threshold
        self._extra = {} if extra is None else dict(extra)

        self._start = None

    def _log(self):
        end = time.time()

        diff = (end - self._start) * 1000
        extra = {
            "time (ms)": diff,
            "start": time.strftime("%X %Z", time.gmtime(self._start)),
            "end": time.strftime("%X %Z", time.gmtime(end)),
            **self._extra
        }
        if self._inline_time:
            message = f"{self._message} execution time is {diff:.2f} ms"
        else:
            message = self._message
        if self._warning_threshold is not None and diff > self._warning_threshold:
            self._logger.warning(message, extra=extra)
        else:
            self._logger.debug(message, extra=extra)


class TimeMeasurer(AbstractTimeMeasurer, ContextManager):
    def __enter__(self):
        self._start = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._log()


class AsyncTimeMeasurer(AbstractTimeMeasurer, AsyncContextManager):
    async def __aenter__(self):
        self._start = time.time()

    async def __aexit__(self, __exc_type, __exc_value, __traceback):
        self._log()


def log_time(f=None, *, logger: Logger = _LOGGER, title: str = None, warning_threshold: float = None):
    def decorator(f):
        fname = (f.__qualname__ if hasattr(f, '__qualname__') else f.__name__) if title is None else title

        if is_async_callable(f):
            @wraps(f)
            async def wrapper(*args, **kwargs):
                async with AsyncTimeMeasurer(
                        fname, inline_time=True, logger=logger, warning_threshold=warning_threshold,
                        extra={'f_args': args, 'f_kwargs': kwargs}
                ):
                    return await f(*args, **kwargs)
        else:
            @wraps(f)
            def wrapper(*args, **kwargs):
                with TimeMeasurer(
                        fname, inline_time=True, logger=logger, warning_threshold=warning_threshold,
                        extra={'f_args': args, 'f_kwargs': kwargs}
                ):
                    return f(*args, **kwargs)

        return wrapper

    if f is not None:
        return decorator(f)
    return decorator



import random
import string
from abc import ABCMeta, abstractmethod
from typing import Generic, TypeVar

from pydantic import ConfigDict, create_model

from tp_interfaces.abstract import AbstractDocumentProcessor, AbstractModelWrapper, ImmutableBaseModel

_Processor = TypeVar('_Processor', bound=AbstractDocumentProcessor)


class WrapperConfig(ImmutableBaseModel):
    def get_processor_config(self) -> ImmutableBaseModel:
        raise NotImplementedError

    model_config = ConfigDict(extra='allow')


_WrapperConfig = TypeVar('_WrapperConfig', bound=WrapperConfig)


class AbstractProcessorWrapper(AbstractModelWrapper[_Processor], Generic[_Processor, _WrapperConfig], metaclass=ABCMeta):

    async def __aenter__(self):
        await super().__aenter__()
        self._config_type = await self._generate_runtime_config()
        return self

    @abstractmethod
    async def _generate_runtime_config(self):
        pass

    @staticmethod
    def _generate_wrapper_config(
        base_config: type[_WrapperConfig],
        model_config: type[ImmutableBaseModel],
        check_intersection: bool = True
    ) -> type[_WrapperConfig]:
        base_fields = base_config.model_fields
        model_fields = model_config.model_fields

        intersection = set(model_fields).intersection(base_fields)
        if check_intersection and intersection:
            raise TypeError(f'filter and model field names collision: {intersection}')

        class _RuntimeConfig(base_config):
            def get_processor_config(self) -> model_config:
                return model_config(**{f: getattr(self, f) for f in model_fields})

        result = create_model(
            f'{base_config.__name__}_{model_config.__name__}_{_id()}',
            __base__=_RuntimeConfig,
            **{name: (field_info.annotation, field_info) for name, field_info in model_fields.items() if name not in base_fields}
        )

        return result


def _id(size=6, chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for _ in range(size))
